@@ -7,8 +7,10 @@ import sqlite3
 import tempfile
 from contextlib import contextmanager
 from inspect import isgeneratorfunction
-from itertools import chain, groupby, product
+from itertools import chain, groupby, product, islice, takewhile, dropwhile
 from pathlib import Path, PurePath
+from collections import deque
+from collections.abc import Iterable
 
 
 class Conn:
@@ -86,47 +88,31 @@ class _InstructionContainer:
         })
         return self
 
-    def _map_group(self, fn):
+    def _map_raw(self, fn):
         pgen = self._genfns[-1]
-        fn = _fn2gen(fn)
 
-        def gen():
-            for _, rs in pgen():
-                yield from fn(rs)
-
-        self._genfns.append(gen)
-        self._insts.append({
-            'cmd': '_map_group',
-            'fn': fn
-        })
-        return self
-
-    def _map_single(self, fn):
-        pgen = self._genfns[-1]
-        fn = _fn2gen(fn)
-
-        def gen():
+        def gen_single():
             for r in pgen():
                 yield from fn(r)
 
-        self._genfns.append(gen)
-        self._insts.append({
-            'cmd': '_map_single',
-            'fn': fn
-        })
-        return self
-
-    # itertools.groupby generates iterator groups, so to turn each into a list
-    def _group2list(self):
-        pgen = self._genfns[-1]
-
-        def gen():
+        def gen_group():
             for k, rs in pgen():
-                yield k, list(rs)
+                yield from fn(k, rs)
+
+        def gen_windowed():
+            for rs in pgen():
+                yield from fn(rs)
+
+        if self._insts[-1]['cmd'] == 'group':
+            self._genfns.append(gen_group)
+        elif self._insts[-1]['cmd'] == 'windowed':
+            self._genfns.append(gen_windowed)
+        else:
+            self._genfns.append(gen_single)
 
-        self._genfns.append(gen)
         self._insts.append({
-            'cmd': '_group2list'
+            'cmd': '_map_raw',
+            'fn': fn
         })
         return self
 
@@ -175,8 +161,65 @@ class _InstructionContainer:
 
     def map(self, fn):
         if self._insts[-1]['cmd'] == 'group':
-            return self._group2list()._map_group(fn)
-        return self._map_single(fn)
+            def _fn2gen(f):
+                # when f is a generator
+                def gen1(_k, rs):
+                    yield from fn(list(rs))
+
+                # when f is an ordinary function
+                def gen2(_k, rs):
+                    x = f(list(rs))
+                    if isinstance(x, dict):
+                        yield x
+                    elif x is not None:
+                        yield from x
+                return gen1 if isgeneratorfunction(f) else gen2
+            return self._map_raw(_fn2gen(fn))
+
+        elif self._insts[-1]['cmd'] == 'windowed':
+            def _fn2gen1(f):
+                # when f is an ordinary function
+                def gen(rs):
+                    x = f(rs)
+                    if isinstance(x, dict):
+                        yield x
+                    elif x is not None:
+                        yield from x
+                return f if isgeneratorfunction(f) else gen
+
+            def _fn2gen2(f):
+                # when f is a generator
+                def gen1(rs):
+                    rs = rs[:step] + [r.copy() for r in rs[step:]]
+                    yield from fn(rs)
+
+                # when f is an ordinary function
+                def gen2(rs):
+                    rs = rs[:step] + [r.copy() for r in rs[step:]]
+                    x = f(rs)
+                    if isinstance(x, dict):
+                        yield x
+                    elif x is not None:
+                        yield from x
+                return gen1 if isgeneratorfunction(f) else gen2
+
+            step = self._insts[-1]['step']
+            n = self._insts[-1]['n']
+
+            if step >= n:
+                # no need to copy dicts
+                return self._map_raw(_fn2gen1(fn))
+            return self._map_raw(_fn2gen2(fn))
+
+        def _fn2gen(f):
+            def gen(r):
+                x = f(r)
+                if isinstance(x, dict):
+                    yield x
+                elif x is not None:
+                    yield from x
+            return f if isgeneratorfunction(f) else gen
+        return self._map_raw(_fn2gen(fn))
 
     def by(self, *cols):
         # need to cut out 'desc', 'asc'
@@ -184,15 +227,11 @@ class _InstructionContainer:
         return self.order(*cols).group(*cols_for_group)
 
     def filter(self, pred):
-        def gen1(r):
+        def gen(r):
             if pred(r):
                 yield r
 
-        def gen2(rs):
-            if pred(rs):
-                yield from rs
-
-        return self.map(gen2 if self._insts[-1]['cmd'] == 'group' else gen1)
+        return self.map(gen)
 
     def update(self, **kwargs):
         def updatefn(r):
@@ -202,14 +241,60 @@ class _InstructionContainer:
 
         return self.map(updatefn)
 
-    def fold(self, **kwargs):
-        def foldfn(rs):
-            r0 = rs[0].copy()
-            for k, v in kwargs.items():
-                r0.update({k: v(rs)})
-            yield r0
+    # pretty expensive for what it actually does
+    def rename(self, **kwargs):
+        pgen = self._genfns[-1]
 
-        return self.map(foldfn)
+        def gen():
+            r0, rs = _spy(pgen())
+            oldcols = list(r0)
+            kwargs_rev = {old: new for new, old in kwargs.items()}
+            newcols = [kwargs_rev.get(old, old) for old in oldcols]
+
+            for r in rs:
+                yield dict(zip(newcols, r.values()))
+
+        self._genfns.append(gen)
+        self._insts.append({
+            'cmd': 'rename',
+            'kwargs': kwargs
+        })
+        return self
+
+    def fold(self, **kwargs):
+        gcols = []
+        if self._insts[-1]['cmd'] == 'group':
+            gcols = self._insts[-1]['cols']
+
+        def foldfn1(rs):
+            r = {}
+            for k, fn in kwargs.items():
+                val = fn(rs)
+                if isinstance(val, list):
+                    for i, v in enumerate(val):
+                        r[k + str(i)] = v
+                else:
+                    r[k] = val
+            yield r
+
+        def foldfn2(keys, rs):
+            rs = list(rs)
+            r = dict(zip(gcols, keys))
+            for k, fn in kwargs.items():
+                val = fn(rs)
+                if not isinstance(val, str) and isinstance(val, Iterable):
+                    for i, v in enumerate(val):
+                        r[k + str(i)] = v
+                else:
+                    r[k] = val
+            yield r
+
+        if self._insts[-1]['cmd'] == 'windowed':
+            return self._map_raw(foldfn1)
+        elif self._insts[-1]['cmd'] == 'group':
+            return self._map_raw(foldfn2)
+        else:
+            raise ValueError("must be grouped or windowed before fold")
 
     def select(self, *cols):
         def selectfn(r):
@@ -226,7 +311,7 @@ class _InstructionContainer:
         return self.map(deselectfn)
 
     def inner_join(self, other):
-        def innerjoin_fn(rs1, rs2, _left, _center, _right):
+        def innerjoin_fn(rs1, rs2, _left, _right):
             if rs1 != [] and rs2 != []:
                 for r1, r2 in product(rs1, rs2):
                     r1.update(r2)
@@ -235,7 +320,7 @@ class _InstructionContainer:
         return self._merge(innerjoin_fn, other)
 
     def left_join(self, other):
-        def leftjoin_fn(rs1, rs2, _left, _center, right):
+        def leftjoin_fn(rs1, rs2, _left, right):
             if rs1 != [] and rs2 != []:
                 for r1, r2 in product(rs1, rs2):
                     r1.update(r2)
@@ -244,13 +329,13 @@ class _InstructionContainer:
             elif rs1 != [] and rs2 == []:
                 for r1 in rs1:
                     for col in right:
-                        r1[col] = None
+                        r1[col] = ''
                     yield r1
 
         return self._merge(leftjoin_fn, other)
 
     def right_join(self, other):
-        def rightjoin_fn(rs1, rs2, left, _center, _right):
+        def rightjoin_fn(rs1, rs2, left, _right):
             if rs1 != [] and rs2 != []:
                 for r1, r2 in product(rs1, rs2):
                     r2.update(r1)
@@ -259,39 +344,95 @@ class _InstructionContainer:
             elif rs1 == [] and rs2 != []:
                 for r2 in rs2:
                     for col in left:
-                        r2[col] = None
+                        r2[col] = ''
                     yield r2
 
         return self._merge(rightjoin_fn, other)
 
     def full_join(self, other):
-        def fulljoin_fn(rs1, rs2, left, _center, right):
+        def fulljoin_fn(rs1, rs2, left, right):
             if rs1 != [] and rs2 != []:
                 for r1, r2 in product(rs1, rs2):
+                    # TODO: updating on the same object multiple times.
+                    # Still, why works?!!,
                     r1.update(r2)
                     yield r1
 
             elif rs1 != [] and rs2 == []:
                 for r1 in rs1:
                     for col in right:
-                        r1[col] = None
+                        r1[col] = ''
                     yield r1
 
             elif rs1 == [] and rs2 != []:
                 for r2 in rs2:
-                    d = {col: None for col in left}
+                    d = {col: '' for col in left}
                     d.update(r2)
                     yield d
 
         return self._merge(fulljoin_fn, other)
 
     def distinct(self, *columns):
-        def distinctfn(rs):
+        def distinctfn(_k, rs):
             yield next(rs)
 
-        return self.by(*columns)._map_group(distinctfn)
+        return self.by(*columns)._map_raw(distinctfn)
+
+    def windowed(self, n, step):
+        pgen = self._genfns[-1]
+
+        def gen():
+            yield from _windowed(pgen(), n, step)
+
+        self._genfns.append(gen)
+        self._insts.append({
+            'cmd': 'windowed',
+            'n': n,
+            'step': step
+        })
+        return self
+
+    def islice(self, *args):
+        pgen = self._genfns[-1]
+
+        def gen():
+            yield from islice(pgen(), *args)
+
+        self._genfns.append(gen)
+        self._insts.append({
+            'cmd': 'islice',
+            'args': args
+        })
+        return self
+
+    def takewhile(self, pred):
+        pgen = self._genfns[-1]
+
+        def gen():
+            yield from takewhile(pred, pgen())
+
+        self._genfns.append(gen)
+        self._insts.append({
+            'cmd': 'takewhile',
+            'pred': pred
+        })
+        return self
+
+    def dropwhile(self, pred):
+        pgen = self._genfns[-1]
+
+        def gen():
+            yield from dropwhile(pred, pgen())
+
+        self._genfns.append(gen)
+        self._insts.append({
+            'cmd': 'dropwhile',
+            'pred': pred
+        })
+        return self
 
     def iter(self):
+        # TODO: analyze self._insts and check if it's semantically ok.
         yield from self._genfns[-1]()
 
     def list(self):
@@ -304,7 +445,7 @@ def _insert_statement(name, d):
     Notice the colons.
     """
     keycols = ', '.join(":" + c.strip() for c in d)
-    return "insert into %s values (%s)" % (name, keycols)
+    return f"insert into {name} values ({keycols})"
 
 
 def _create_statement(tname, cols):
@@ -314,7 +455,7 @@ def _create_statement(tname, cols):
         Every type is numeric.
     """
     schema = ', '.join([col + ' ' + 'numeric' for col in cols])
-    return "create table if not exists %s (%s)" % (tname, schema)
+    return f"create table if not exists {tname} ({schema})"
 
 
 def _dict_factory(cursor, row):
@@ -358,18 +499,6 @@ def _fetch(c, tname, cols):
     yield from c.cursor().execute(query)
 
 
-# ordinary function to generator function
-def _fn2gen(f):
-    def gen(r):
-        x = f(r)
-        if isinstance(x, dict):
-            yield x
-        elif x is not None:
-            yield from x
-
-    return f if isgeneratorfunction(f) else gen
-
-
 def _spy(it):
     val = next(it)
     return val, chain([val], it)
@@ -385,40 +514,35 @@ def _step(fn, krs1, krs2):
         r1, rs1 = _spy(rs1)
         r2, rs2 = _spy(rs2)
 
-        cols1 = list(r1)
-        cols2 = list(r2)
-
-        left = [c for c in cols1 if c not in cols2]
-        # common columns
-        center = [c for c in cols1 if c in cols2]
-        right = [c for c in cols2 if c not in cols1]
+        left = [c for c in list(r1) if c not in list(r2)]
+        right = [c for c in list(r2) if c not in list(r1)]
 
         while True:
             if k1 == k2:
-                yield from fn(rs1, rs2, left, center, right)
+                yield from fn(rs1, rs2, left, right)
                 k1 = k2 = empty
                 k1, rs1 = next(krs1)
                 k2, rs2 = next(krs2)
             elif k1 < k2:
-                yield from fn(rs1, [], left, center, right)
+                yield from fn(rs1, [], left, right)
                 k1 = empty
                 k1, rs1 = next(krs1)
             else:
-                yield from fn([], rs2, left, center, right)
+                yield from fn([], rs2, left, right)
                 k2 = empty
                 k2, rs2 = next(krs2)
 
     except StopIteration:
         # unconsumed
         if k1 is not empty:
-            yield from fn(rs1, [], left, center, right)
+            yield from fn(rs1, [], left, right)
         if k2 is not empty:
-            yield from fn([], rs2, left, center, right)
+            yield from fn([], rs2, left, right)
 
         for _, rs1 in krs1:
-            yield from fn(rs1, [], left, center, right)
+            yield from fn(rs1, [], left, right)
         for _, rs2 in krs2:
-            yield from fn([], rs2, left, center, right)
+            yield from fn([], rs2, left, right)
 
 
 def _inst2iter(obj):
@@ -437,6 +561,32 @@ def _build_initgen(conn, tname, cols):
     return initgen
 
 
+def _windowed(seq, n, step):
+    if n < 0:
+        raise ValueError('n must be >= 0')
+    if n == 0:
+        yield []
+        return
+    if step < 1:
+        raise ValueError('step must be >= 1')
+
+    window = deque(maxlen=n)
+    i = n
+    for _ in map(window.append, seq):
+        i -= 1
+        if not i:
+            i = step
+            yield list(window)
+
+    size = len(window)
+    if size == 0:
+        return
+    if size < n:
+        yield list(window)
+    elif 0 < i < min(step, n):
+        yield list(window)[i:]
+
+
 @contextmanager
 def _connect(db):
     conn = sqlite3.connect(db)
